import JSBI from 'jsbi';
import { fromByteArray } from 'base64-js';

/**
 * ExtData is used to handle Extension Types that are not registered to ExtensionCodec.
 */
class ExtData {
    constructor(type, data) {
        this.type = type;
        this.data = data;
        this.type = type;
        this.data = data;
    }
}

// ExtensionCodec to handle MessagePack extensions
class ExtensionCodec {
    constructor() {
        // custom extensions
        this.encoders = [];
        this.decoders = [];
        this.identifiers = [];
    }
    register({ type, identifier, encode, decode, }) {
        this.encoders[type] = encode;
        this.decoders[type] = decode;
        this.identifiers.push(identifier);
    }
    encode(object) {
        let type = null;
        for (let i = 0; i < this.identifiers.length; i++) {
            const id = this.identifiers[i];
            if (id(object)) {
                type = i;
            }
        }
        if (type === null) {
            throw new Error(`Unrecognized object: ${Object.prototype.toString.apply(object)}`);
        }
        const encoder = this.encoders[type];
        const data = encoder(object);
        return new ExtData(type, data);
    }
    decode(data, type) {
        const decoder = this.decoders[type];
        return decoder(data, type);
    }
}
ExtensionCodec.defaultCodec = new ExtensionCodec();

/**
 * Check if a value looks like a JSBI value.
 */
function isJsbi(value) {
    // if it has __clzmsd it's probably jsbi
    return Array.isArray(value) && '__clzmsd' in value;
}
function isPlainObject(object) {
    if (Object.getPrototypeOf(object) === null) {
        return true;
    }
    let proto = object;
    while (Object.getPrototypeOf(proto) !== null) {
        proto = Object.getPrototypeOf(proto);
    }
    return Object.getPrototypeOf(object) === proto;
}

// Copyright (c) 2018, Arista Networks, Inc.
class Float32 {
    /**
     * A class to represent an explicit float32
     */
    constructor(value) {
        this.type = Float32.type;
        this.value = value ? parseFloat(String(value)) : 0.0; // undefined defaults to 0.0
    }
    toString() {
        const strValue = this.value.toString();
        const hasDecimal = strValue.includes('.');
        return hasDecimal ? strValue : this.value.toFixed(1);
    }
}
Float32.type = 'float32';
class Float64 {
    /**
     * A class to represent an explicit float64
     */
    constructor(value) {
        this.type = Float64.type;
        this.value = value ? parseFloat(String(value)) : 0.0; // undefined defaults to 0.0
    }
    toString() {
        const strValue = this.value.toString();
        const hasDecimal = strValue.includes('.');
        return hasDecimal ? strValue : this.value.toFixed(1);
    }
}
Float64.type = 'float64';
class Int {
    /**
     * A class to represent an explicit int
     */
    constructor(value, forceJSBI = false) {
        this.type = Int.type;
        if (typeof value === 'string') {
            let BI; // This is a type alias for JSBI/BigInt
            if (typeof BigInt === 'undefined' || forceJSBI) {
                BI = JSBI.BigInt;
            }
            else {
                BI = BigInt;
            }
            this.value =
                parseInt(value, 10) > 0xffffffff || parseInt(value, 10) < -0x80000000
                    ? BI(value) // eslint-disable-line new-cap
                    : parseInt(value, 10);
        }
        else if (typeof value === 'bigint' || isJsbi(value)) {
            this.value = value;
        }
        else {
            this.value = parseInt(String(value), 10);
        }
    }
    toString() {
        return this.value.toString();
    }
}
Int.type = 'int';
class Bool {
    /**
     * A class to represent an explicit boolean
     */
    constructor(value) {
        this.type = Bool.type;
        this.value = !!value;
    }
    toString() {
        return this.value ? 'true' : 'false';
    }
}
Bool.type = 'bool';
class Nil {
    /**
     * A class to represent an explicit Nil
     */
    constructor() {
        this.type = Nil.type;
        this.value = null;
    }
    toString() {
        return 'null';
    }
}
Nil.type = 'nil';
class Str {
    /**
     * A class to represent an explicit String
     */
    constructor(value) {
        this.type = Str.type;
        switch (typeof value) {
            case 'string':
                this.value = value;
                break;
            case 'bigint':
                this.value = value.toString();
                break;
            case 'undefined':
                this.value = '';
                break;
            default:
                this.value = JSON.stringify(value);
        }
        if (isJsbi(value)) {
            this.value = value.toString();
        }
    }
    toString() {
        return this.value;
    }
}
Str.type = 'str';
class Pointer {
    /**
     * A class to represent a Pointer type.
     * A Pointer is a pointer from one set of path elements to another.
     */
    constructor(value = []) {
        this.delimiter = '/';
        this.type = Pointer.type;
        let strValue;
        if (!Array.isArray(value)) {
            if (typeof value !== 'string') {
                strValue = JSON.stringify(value);
            }
            else {
                strValue = value;
            }
            const ptrArray = strValue.split(this.delimiter);
            while (ptrArray[0] === '') {
                ptrArray.shift();
            }
            this.value = ptrArray.map((pathEl) => {
                try {
                    return JSON.parse(pathEl);
                }
                catch (e) {
                    // ignore errors, these are just regular strings
                }
                return pathEl;
            });
        }
        else {
            this.value = value;
        }
    }
    toString() {
        return this.value
            .map((pathEl) => {
            if (typeof pathEl === 'string') {
                return pathEl;
            }
            return JSON.stringify(pathEl);
        })
            .join(this.delimiter);
    }
}
Pointer.type = 'ptr';
class Wildcard {
    /**
     * A class to represent a Wildcard type.
     * A Wildcard is a type that matches 1 or more path elements
     */
    constructor() {
        this.type = Wildcard.type;
        this.value = null;
    }
    toString() {
        return '*';
    }
}
Wildcard.type = '*';

/* eslint-disable @typescript-eslint/naming-convention */
function isFloat32(value) {
    return value.type === 'float32';
}
function isFloat64(value) {
    return value.type === 'float64';
}
function isInt(value) {
    return value.type === 'int';
}
function isStr(value) {
    return value.type === 'str';
}
function isBool(value) {
    return value.type === 'bool';
}
function isBasicNeatType(value) {
    if (typeof value === 'object' && value !== null) {
        if (Object.keys(value).length === 2 && 'type' in value && 'value' in value) {
            const neatValue = value;
            return [Bool.type, Float32.type, Float64.type, Int.type, Nil.type, Str.type].includes(neatValue.type);
        }
    }
    return false;
}
function isNeatType(value) {
    if (typeof value === 'object' && value !== null) {
        if (isBasicNeatType(value)) {
            return true;
        }
        if (Object.keys(value).length === 3 &&
            'type' in value &&
            'value' in value &&
            'delimiter' in value) {
            const neatValue = value;
            return neatValue.type === Pointer.type;
        }
    }
    return false;
}
function sortByBinaryValue(a, b) {
    const len = Math.min(a.length, b.length);
    if (a === b) {
        return 0;
    }
    for (let i = 0; i < len; i += 1) {
        if (a[i] !== b[i]) {
            return a[i] - b[i];
        }
    }
    return a.length - b.length;
}
function sortMapByKey(a, b) {
    const aVal = a[0];
    const bVal = b[0];
    return sortByBinaryValue(aVal, bVal);
}
class NeatTypeSerializer {
    static serialize(neatTypeInstance) {
        return {
            __neatTypeClass: neatTypeInstance.type,
            value: neatTypeInstance.value,
        };
    }
    static deserialize(serializedNeatType) {
        // eslint-disable-next-line no-underscore-dangle
        return new NeatTypeSerializer.NEAT_TYPE_MAP[serializedNeatType.__neatTypeClass](serializedNeatType.value);
    }
}
NeatTypeSerializer.NEAT_TYPE_MAP = {
    float32: Float32,
    float64: Float64,
    int: Int,
    ptr: Pointer,
    str: Str,
    nil: Nil,
    bool: Bool,
};

// DataView extension to handle int64 / uint64,
// where the actual range is 53-bits integer (a.k.a. safe integer)
function bufferToString(high, low, unsigned) {
    let highNum = high;
    let lowNum = low;
    const radix = 10;
    const sign = !unsigned && high & 0x80000000;
    if (sign) {
        highNum = ~high;
        lowNum = 4294967296 - low;
    }
    let str = '';
    // eslint-disable-next-line no-constant-condition
    while (true) {
        const mod = (highNum % radix) * 4294967296 + lowNum;
        highNum = Math.floor(highNum / radix);
        lowNum = Math.floor(mod / radix);
        str = (mod % radix).toString(radix) + str;
        if (!highNum && !lowNum) {
            break;
        }
    }
    if (sign) {
        str = '-' + str;
    }
    return str;
}
function fromStringToBuffer(view, offset, str) {
    const radix = 10;
    let pos = 0;
    const len = str.length;
    let highNum = 0;
    let lowNum = 0;
    if (str.startsWith('-')) {
        pos++;
    }
    const sign = pos;
    while (pos < len) {
        const chr = parseInt(str[pos++], radix);
        if (Number.isNaN(chr) || chr < 0) {
            // NaN
            break;
        }
        lowNum = lowNum * radix + chr;
        highNum = highNum * radix + Math.floor(lowNum / 4294967296);
        lowNum %= 4294967296;
    }
    if (sign) {
        highNum = ~highNum;
        if (lowNum) {
            lowNum = 4294967296 - lowNum;
        }
        else {
            highNum++;
        }
    }
    view.setUint32(offset, highNum);
    view.setUint32(offset + 4, lowNum);
}
function setUint64(view, offset, value) {
    if (typeof value === 'number') {
        const high = value / 4294967296;
        const low = value; // high bits are truncated by DataView
        view.setUint32(offset, high);
        view.setUint32(offset + 4, low);
    }
    else if (typeof value === 'bigint') {
        view.setBigUint64(offset, value);
    }
    else if (typeof value === 'string') {
        fromStringToBuffer(view, offset, value);
    }
}
function setInt64(view, offset, value) {
    if (typeof value === 'number') {
        const high = Math.floor(value / 4294967296);
        const low = value; // high bits are truncated by DataView
        view.setUint32(offset, high);
        view.setUint32(offset + 4, low);
    }
    else if (typeof value === 'bigint') {
        view.setBigInt64(offset, value);
    }
    else if (typeof value === 'string') {
        fromStringToBuffer(view, offset, value);
    }
}
function getInt64(view, offset, useJSBI) {
    if (!view.getBigInt64 || useJSBI) {
        const high = view.getInt32(offset);
        const low = view.getUint32(offset + 4);
        return JSBI.BigInt(bufferToString(high, low, false));
    }
    return view.getBigInt64(offset);
}
function getUint64(view, offset, useJSBI) {
    if (!view.getBigUint64 || useJSBI) {
        const high = view.getUint32(offset);
        const low = view.getUint32(offset + 4);
        return JSBI.BigInt(bufferToString(high, low, true));
    }
    return view.getBigUint64(offset);
}

function ensureUint8Array(buffer) {
    if (buffer instanceof Uint8Array) {
        return buffer;
    }
    if (ArrayBuffer.isView(buffer)) {
        return new Uint8Array(buffer.buffer, buffer.byteOffset, buffer.byteLength);
    }
    if (buffer instanceof ArrayBuffer) {
        return new Uint8Array(buffer);
    }
    return Uint8Array.from(buffer);
}
function createDataView(buffer) {
    if (buffer instanceof ArrayBuffer) {
        return new DataView(buffer);
    }
    const bufferView = ensureUint8Array(buffer);
    return new DataView(bufferView.buffer, bufferView.byteOffset, bufferView.byteLength);
}

const TEXT_ENCODING_AVAILABLE = process.env.TEXT_ENCODING !== 'never' &&
    typeof TextEncoder !== 'undefined' &&
    typeof TextDecoder !== 'undefined';
function utf8Count(str) {
    const strLength = str.length;
    let byteLength = 0;
    let pos = 0;
    while (pos < strLength) {
        let value = str.charCodeAt(pos++);
        if ((value & 0xffffff80) === 0) {
            // 1-byte
            byteLength++;
            continue;
        }
        else if ((value & 0xfffff800) === 0) {
            // 2-bytes
            byteLength += 2;
        }
        else {
            // handle surrogate pair
            if (value >= 0xd800 && value <= 0xdbff) {
                // high surrogate
                if (pos < strLength) {
                    const extra = str.charCodeAt(pos);
                    if ((extra & 0xfc00) === 0xdc00) {
                        ++pos;
                        value = ((value & 0x3ff) << 10) + (extra & 0x3ff) + 0x10000;
                    }
                }
            }
            if ((value & 0xffff0000) === 0) {
                // 3-byte
                byteLength += 3;
            }
            else {
                // 4-byte
                byteLength += 4;
            }
        }
    }
    return byteLength;
}
function utf8EncodeJs(str, output, outputOffset) {
    const strLength = str.length;
    let offset = outputOffset;
    let pos = 0;
    while (pos < strLength) {
        let value = str.charCodeAt(pos++);
        if ((value & 0xffffff80) === 0) {
            // 1-byte
            output[offset++] = value;
            continue;
        }
        else if ((value & 0xfffff800) === 0) {
            // 2-bytes
            output[offset++] = ((value >> 6) & 0x1f) | 0xc0;
        }
        else {
            // handle surrogate pair
            if (value >= 0xd800 && value <= 0xdbff) {
                // high surrogate
                if (pos < strLength) {
                    const extra = str.charCodeAt(pos);
                    if ((extra & 0xfc00) === 0xdc00) {
                        ++pos;
                        value = ((value & 0x3ff) << 10) + (extra & 0x3ff) + 0x10000;
                    }
                }
            }
            if ((value & 0xffff0000) === 0) {
                // 3-byte
                output[offset++] = ((value >> 12) & 0x0f) | 0xe0;
                output[offset++] = ((value >> 6) & 0x3f) | 0x80;
            }
            else {
                // 4-byte
                output[offset++] = ((value >> 18) & 0x07) | 0xf0;
                output[offset++] = ((value >> 12) & 0x3f) | 0x80;
                output[offset++] = ((value >> 6) & 0x3f) | 0x80;
            }
        }
        output[offset++] = (value & 0x3f) | 0x80;
    }
}
const sharedTextEncoder = TEXT_ENCODING_AVAILABLE ? new TextEncoder() : undefined;
const TEXT_ENCODER_THRESHOLD = process.env.TEXT_ENCODING !== 'force' ? 200 : 0;
function utf8EncodeTEencode(str, output, outputOffset) {
    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    output.set(sharedTextEncoder.encode(str), outputOffset);
}
function utf8EncodeTEencodeInto(str, output, outputOffset) {
    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    sharedTextEncoder.encodeInto(str, output.subarray(outputOffset));
}
const utf8EncodeTE = (sharedTextEncoder === null || sharedTextEncoder === void 0 ? void 0 : sharedTextEncoder.encodeInto)
    ? utf8EncodeTEencodeInto
    : utf8EncodeTEencode;
const CHUNK_SIZE = 4096;
function utf8DecodeJs(bytes, inputOffset, byteLength) {
    let offset = inputOffset;
    const end = offset + byteLength;
    const units = [];
    let result = '';
    while (offset < end) {
        const byte1 = bytes[offset++];
        if ((byte1 & 0x80) === 0) {
            // 1 byte
            units.push(byte1);
        }
        else if ((byte1 & 0xe0) === 0xc0) {
            // 2 bytes
            const byte2 = bytes[offset++] & 0x3f;
            units.push(((byte1 & 0x1f) << 6) | byte2);
        }
        else if ((byte1 & 0xf0) === 0xe0) {
            // 3 bytes
            const byte2 = bytes[offset++] & 0x3f;
            const byte3 = bytes[offset++] & 0x3f;
            units.push(((byte1 & 0x1f) << 12) | (byte2 << 6) | byte3);
        }
        else if ((byte1 & 0xf8) === 0xf0) {
            // 4 bytes
            const byte2 = bytes[offset++] & 0x3f;
            const byte3 = bytes[offset++] & 0x3f;
            const byte4 = bytes[offset++] & 0x3f;
            let unit = ((byte1 & 0x07) << 0x12) | (byte2 << 0x0c) | (byte3 << 0x06) | byte4;
            if (unit > 0xffff) {
                unit -= 0x10000;
                units.push(((unit >>> 10) & 0x3ff) | 0xd800);
                unit = 0xdc00 | (unit & 0x3ff);
            }
            units.push(unit);
        }
        else {
            units.push(byte1);
        }
        if (units.length >= CHUNK_SIZE) {
            result += String.fromCharCode(...units);
            units.length = 0;
        }
    }
    if (units.length > 0) {
        result += String.fromCharCode(...units);
    }
    return result;
}
const sharedTextDecoder = TEXT_ENCODING_AVAILABLE ? new TextDecoder() : null;
const TEXT_DECODER_THRESHOLD = process.env.TEXT_DECODER !== 'force' ? 200 : 0;
function utf8DecodeTD(bytes, inputOffset, byteLength) {
    const stringBytes = bytes.subarray(inputOffset, inputOffset + byteLength);
    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    return sharedTextDecoder.decode(stringBytes);
}

const DEFAULT_MAX_DEPTH = 100;
const DEFAULT_INITIAL_BUFFER_SIZE = 2048;
class Encoder {
    constructor(extensionCodec = ExtensionCodec.defaultCodec, maxDepth = DEFAULT_MAX_DEPTH, initialBufferSize = DEFAULT_INITIAL_BUFFER_SIZE) {
        this.pos = 0;
        this.extensionCodec = extensionCodec;
        this.maxDepth = maxDepth;
        this.initialBufferSize = initialBufferSize;
        this.view = new DataView(new ArrayBuffer(this.initialBufferSize));
        this.bytes = new Uint8Array(this.view.buffer);
    }
    keyEncoder(value, depth) {
        const encoder = new Encoder(this.extensionCodec, this.maxDepth, this.initialBufferSize);
        encoder.encode(value, depth);
        return encoder.getUint8Array();
    }
    encode(object, depth) {
        if (depth > this.maxDepth) {
            throw new Error(`Too deep objects in depth ${depth}`);
        }
        if (object == null) {
            this.encodeNil();
        }
        else if (typeof object === 'boolean') {
            this.encodeBoolean(object);
        }
        else if (typeof object === 'number') {
            this.encodeNumber(object);
        }
        else if (typeof object === 'string') {
            this.encodeString(object);
        }
        else {
            this.encodeObject(object, depth);
        }
    }
    getUint8Array() {
        return this.bytes.subarray(0, this.pos);
    }
    ensureBufferSizeToWrite(sizeToWrite) {
        const requiredSize = this.pos + sizeToWrite;
        if (this.view.byteLength < requiredSize) {
            this.resizeBuffer(requiredSize * 2);
        }
    }
    resizeBuffer(newSize) {
        const newBuffer = new ArrayBuffer(newSize);
        const newBytes = new Uint8Array(newBuffer);
        const newView = new DataView(newBuffer);
        newBytes.set(this.bytes);
        this.view = newView;
        this.bytes = newBytes;
    }
    encodeNil() {
        this.writeU8(0xc0);
    }
    encodeBoolean(object) {
        if (object === false) {
            this.writeU8(0xc2);
        }
        else {
            this.writeU8(0xc3);
        }
    }
    encodeNumber(object) {
        if (Number.isSafeInteger(object)) {
            if (object >= 0) {
                if (object < 0x80) {
                    // positive fixint
                    this.writeU8(object);
                }
                else if (object < 0x100) {
                    // uint 8
                    this.writeU8(0xcc);
                    this.writeU8(object);
                }
                else if (object < 0x10000) {
                    // uint 16
                    this.writeU8(0xcd);
                    this.writeU16(object);
                }
                else if (object < 0x100000000) {
                    // uint 32
                    this.writeU8(0xce);
                    this.writeU32(object);
                }
                else {
                    // uint 64
                    this.writeU8(0xcf);
                    this.writeU64(object);
                }
            }
            else if (object >= -0x20) {
                // negative fixint
                this.writeU8(0xe0 | (object + 0x20));
            }
            else if (object >= -0x80) {
                // int 8
                this.writeU8(0xd0);
                this.writeI8(object);
            }
            else if (object >= -0x8000) {
                // int 16
                this.writeU8(0xd1);
                this.writeI16(object);
            }
            else if (object >= -0x80000000) {
                // int 32
                this.writeU8(0xd2);
                this.writeI32(object);
            }
            else {
                // int 64
                this.writeU8(0xd3);
                this.writeI64(object);
            }
        }
        else {
            // non-integer numbers
            // float 64
            this.writeU8(0xcb);
            this.writeF64(object);
        }
    }
    writeStringHeader(size) {
        if (size < 0x100) {
            // bin 8
            this.writeU8(0xc4);
            this.writeU8(size);
        }
        else if (size < 0x10000) {
            // bin 16
            this.writeU8(0xc5);
            this.writeU16(size);
        }
        else {
            // bin 32
            this.writeU8(0xc6);
            this.writeU32(size);
        }
    }
    encodeString(object) {
        const maxHeaderSize = 1 + 4;
        const strLength = object.length;
        if (TEXT_ENCODING_AVAILABLE && strLength > TEXT_ENCODER_THRESHOLD) {
            const byteLength = utf8Count(object);
            this.ensureBufferSizeToWrite(maxHeaderSize + byteLength);
            this.writeStringHeader(byteLength);
            utf8EncodeTE(object, this.bytes, this.pos);
            this.pos += byteLength;
        }
        else {
            const byteLength = utf8Count(object);
            this.ensureBufferSizeToWrite(maxHeaderSize + byteLength);
            this.writeStringHeader(byteLength);
            utf8EncodeJs(object, this.bytes, this.pos);
            this.pos += byteLength;
        }
    }
    encodeObject(object, depth) {
        if (Array.isArray(object)) {
            if (isJsbi(object)) {
                this.encodeJSBI(object);
            }
            else {
                this.encodeArray(object, depth);
            }
        }
        else if (ArrayBuffer.isView(object)) {
            this.encodeBinary(object);
        }
        else if (typeof object === 'bigint') {
            this.encodeBigInt(object);
        }
        else if (object instanceof Map) {
            this.encodeMap(object, depth);
        }
        else if (typeof object === 'object') {
            if (isBasicNeatType(object)) {
                this.encodeNeatClass(object);
            }
            else if (isPlainObject(object)) {
                // Find out if it is a plain object
                this.encodePlainObject(object, depth);
            }
            else {
                // Otherwise try to encode objects with custom codec of non-primitives
                this.encodeExtension(this.extensionCodec.encode(object));
            }
        }
        else if (typeof object === 'function') {
            this.encodeNil();
        }
        else {
            // symbol, function and other special object come here unless extensionCodec handles them.
            throw new Error(`Unrecognized object: ${Object.prototype.toString.apply(object)}`);
        }
    }
    encodeNeatClass(value) {
        if (isFloat32(value)) {
            // float 32 -- 0xca
            this.writeU8(0xca);
            this.writeF32(value.value);
        }
        else if (isFloat64(value)) {
            // float 64 -- 0xcb
            this.writeU8(0xcb);
            this.writeF64(value.value);
        }
        else if (isInt(value)) {
            // int
            if (typeof value.value === 'bigint') {
                this.encodeBigInt(value.value);
            }
            else if (isJsbi(value.value)) {
                this.encodeJSBI(value.value);
            }
            else {
                this.encodeNumber(value.value);
            }
        }
        else if (isStr(value)) {
            // string
            this.encodeString(value.value);
        }
        else if (isBool(value)) {
            // bool
            this.encodeBoolean(value.value);
        }
        else {
            // nil
            this.encodeNil();
        }
    }
    encodeBigInt(value) {
        // uint 64 -- 0xcf
        // int 64 -- 0xd3
        if (value === BigInt(0) ||
            BigInt.asIntN(32, value) > 0 ||
            (value < 0 && BigInt.asUintN(32, value) === value * BigInt(-1))) {
            this.encodeNumber(Number(value));
        }
        else if (value < 0) {
            this.writeU8(0xd3);
            this.writeI64(value);
        }
        else if (value < 0x100000000) {
            this.writeU8(0xce);
            this.writeU32(Number(value));
        }
        else {
            this.writeU8(0xcf);
            this.writeU64(value);
        }
    }
    encodeJSBI(value) {
        const strValue = value.toString();
        if (strValue === '0' ||
            (JSBI.LE(value, 0x7fffffff) && JSBI.GE(value, -0xffffffff)) ||
            (strValue < '0' && JSBI.GE(value, -0xffffffff))) {
            this.encodeNumber(JSBI.toNumber(value));
        }
        else if (strValue < '0') {
            this.writeU8(0xd3);
            this.writeI64(strValue);
        }
        else if (JSBI.LE(value, 0xffffffff)) {
            this.writeU8(0xce);
            this.writeU32(JSBI.toNumber(value));
        }
        else {
            this.writeU8(0xcf);
            this.writeU64(strValue);
        }
    }
    encodeBinary(object) {
        const size = object.byteLength;
        if (size < 0x100) {
            // bin 8
            this.writeU8(0xc4);
            this.writeU8(size);
        }
        else if (size < 0x10000) {
            // bin 16
            this.writeU8(0xc5);
            this.writeU16(size);
        }
        else {
            // bin 32
            this.writeU8(0xc6);
            this.writeU32(size);
        }
        const bytes = ensureUint8Array(object);
        this.writeU8a(bytes);
    }
    encodeArray(object, depth) {
        const size = object.length;
        if (size < 16) {
            // fixarray
            this.writeU8(0x90 + size);
        }
        else if (size < 0x10000) {
            // array 16
            this.writeU8(0xdc);
            this.writeU16(size);
        }
        else {
            // array 32
            this.writeU8(0xdd);
            this.writeU32(size);
        }
        for (const item of object) {
            this.encode(item, depth + 1);
        }
    }
    encodeMap(object, depth) {
        const size = object.size;
        const sortedMap = [];
        if (size < 16) {
            // fixmap
            this.writeU8(0x80 + size);
        }
        else if (size < 0x10000) {
            // map 16
            this.writeU8(0xde);
            this.writeU16(size);
        }
        else {
            // map 32
            this.writeU8(0xdf);
            this.writeU32(size);
        }
        object.forEach((value, key) => {
            const encodedKey = this.keyEncoder(key, depth + 1);
            sortedMap.push([encodedKey, value]);
        });
        sortedMap.sort(sortMapByKey);
        sortedMap.forEach((keyVal) => {
            this.writeU8a(keyVal[0]);
            this.encode(keyVal[1], depth + 1);
        });
    }
    encodePlainObject(object, depth) {
        this.encodeMap(new Map(Object.entries(object)), depth);
    }
    encodeExtension(ext) {
        const size = ext.data.length;
        if (size === 1) {
            // fixext 1
            this.writeU8(0xd4);
        }
        else if (size === 2) {
            // fixext 2
            this.writeU8(0xd5);
        }
        else if (size === 4) {
            // fixext 4
            this.writeU8(0xd6);
        }
        else if (size === 8) {
            // fixext 8
            this.writeU8(0xd7);
        }
        else if (size === 16) {
            // fixext 16
            this.writeU8(0xd8);
        }
        else if (size < 0x100) {
            // ext 8
            this.writeU8(0xc7);
            this.writeU8(size);
        }
        else if (size < 0x10000) {
            // ext 16
            this.writeU8(0xc8);
            this.writeU16(size);
        }
        else {
            // ext 32
            this.writeU8(0xc9);
            this.writeU32(size);
        }
        this.writeI8(ext.type);
        this.writeU8a(ext.data);
    }
    writeU8(value) {
        this.ensureBufferSizeToWrite(1);
        this.view.setUint8(this.pos, value);
        this.pos++;
    }
    writeU8a(values) {
        const size = values.length;
        this.ensureBufferSizeToWrite(size);
        this.bytes.set(values, this.pos);
        this.pos += size;
    }
    writeI8(value) {
        this.ensureBufferSizeToWrite(1);
        this.view.setInt8(this.pos, value);
        this.pos++;
    }
    writeU16(value) {
        this.ensureBufferSizeToWrite(2);
        this.view.setUint16(this.pos, value);
        this.pos += 2;
    }
    writeI16(value) {
        this.ensureBufferSizeToWrite(2);
        this.view.setInt16(this.pos, value);
        this.pos += 2;
    }
    writeU32(value) {
        this.ensureBufferSizeToWrite(4);
        this.view.setUint32(this.pos, value);
        this.pos += 4;
    }
    writeI32(value) {
        this.ensureBufferSizeToWrite(4);
        this.view.setInt32(this.pos, value);
        this.pos += 4;
    }
    writeF32(value) {
        this.ensureBufferSizeToWrite(4);
        this.view.setFloat32(this.pos, value);
        this.pos += 4;
    }
    writeF64(value) {
        this.ensureBufferSizeToWrite(8);
        this.view.setFloat64(this.pos, value);
        this.pos += 8;
    }
    writeU64(value) {
        this.ensureBufferSizeToWrite(8);
        setUint64(this.view, this.pos, value);
        this.pos += 8;
    }
    writeI64(value) {
        this.ensureBufferSizeToWrite(8);
        setInt64(this.view, this.pos, value);
        this.pos += 8;
    }
}

const defaultEncodeOptions = {};
/**
 * It encodes `value` in the MessagePack format and
 * returns a byte buffer.
 *
 * The returned buffer is a slice of a larger `ArrayBuffer`, so you have to use its `#byteOffset`
 * and `#byteLength` in order to convert it to another typed arrays including NodeJS `Buffer`.
 */
function encode(value, options = defaultEncodeOptions) {
    const encoder = new Encoder(options.extensionCodec, options.maxDepth, options.initialBufferSize);
    encoder.encode(value, 1);
    return encoder.getUint8Array();
}

/* eslint-disable no-labels */
const DEFAULT_MAX_KEY_LENGTH = 16;
const DEFAULT_MAX_LENGTH_PER_KEY = 16;
class CachedKeyDecoder {
    constructor(maxKeyLength = DEFAULT_MAX_KEY_LENGTH, maxLengthPerKey = DEFAULT_MAX_LENGTH_PER_KEY) {
        this.maxKeyLength = maxKeyLength;
        this.maxLengthPerKey = maxLengthPerKey;
        // avoid `new Array(N)` to create a non-sparse array for performance.
        this.caches = [];
        for (let i = 0; i < this.maxKeyLength; i++) {
            this.caches.push([]);
        }
    }
    canBeCached(byteLength) {
        return byteLength > 0 && byteLength <= this.maxKeyLength;
    }
    get(bytes, inputOffset, byteLength) {
        const records = this.caches[byteLength - 1];
        const recordsLength = records.length;
        FIND_CHUNK: for (let i = 0; i < recordsLength; i++) {
            const record = records[i];
            const recordBytes = record.bytes;
            for (let j = 0; j < byteLength; j++) {
                if (recordBytes[j] !== bytes[inputOffset + j]) {
                    continue FIND_CHUNK;
                }
            }
            return record.value;
        }
        return null;
    }
    store(bytes, value) {
        const records = this.caches[bytes.length - 1];
        const record = { bytes, value };
        if (records.length >= this.maxLengthPerKey) {
            // `records` are full!
            // Set `record` to a randomized position.
            records[(Math.random() * records.length) | 0] = record;
        }
        else {
            records.push(record);
        }
    }
    decode(bytes, inputOffset, byteLength) {
        const cachedValue = this.get(bytes, inputOffset, byteLength);
        if (cachedValue) {
            return cachedValue;
        }
        const value = utf8DecodeJs(bytes, inputOffset, byteLength);
        // Ensure to copy a slice of bytes because the byte may be NodeJS Buffer and Buffer#slice()
        // returns a reference to its internal ArrayBuffer.
        const slicedCopyOfBytes = Uint8Array.prototype.slice.call(bytes, inputOffset, inputOffset + byteLength);
        this.store(slicedCopyOfBytes, value);
        return value;
    }
}

function prettyByte(byte) {
    return `${byte < 0 ? '-' : ''}0x${Math.abs(byte).toString(16).padStart(2, '0')}`;
}

/* eslint-disable no-labels, no-extra-label */
const isValidMapKeyType = (key) => {
    const keyType = typeof key;
    return keyType === 'string';
};
const EMPTY_VIEW = new DataView(new ArrayBuffer(0));
const EMPTY_BYTES = new Uint8Array(EMPTY_VIEW.buffer);
const DEFAULT_MAX_LENGTH = 4294967295; // uint32_max
const sharedCachedKeyDecoder = new CachedKeyDecoder();
class Decoder {
    constructor(extensionCodec = ExtensionCodec.defaultCodec, useJSBI = false, maxStrLength = DEFAULT_MAX_LENGTH, maxBinLength = DEFAULT_MAX_LENGTH, maxArrayLength = DEFAULT_MAX_LENGTH, maxMapLength = DEFAULT_MAX_LENGTH, maxExtLength = DEFAULT_MAX_LENGTH, cachedKeyDecoder = sharedCachedKeyDecoder) {
        this.totalPos = 0;
        this.pos = 0;
        this.view = EMPTY_VIEW;
        this.bytes = EMPTY_BYTES;
        this.stack = [];
        this.extensionCodec = extensionCodec;
        this.cachedKeyDecoder = cachedKeyDecoder;
        this.useJSBI = useJSBI;
        this.maxStrLength = maxStrLength;
        this.maxBinLength = maxBinLength;
        this.maxArrayLength = maxArrayLength;
        this.maxMapLength = maxMapLength;
        this.maxExtLength = maxExtLength;
    }
    setBuffer(buffer) {
        this.bytes = ensureUint8Array(buffer);
        this.view = createDataView(this.bytes);
        this.pos = 0;
    }
    hasRemaining(size = 1) {
        return this.view.byteLength - this.pos >= size;
    }
    createNoExtraBytesError(posToShow) {
        const { view, pos } = this;
        return new RangeError(`Extra ${view.byteLength - pos} byte(s) found at buffer[${posToShow}]`);
    }
    decodeSingleSync() {
        const object = this.decodeSync();
        if (this.hasRemaining()) {
            throw this.createNoExtraBytesError(this.pos);
        }
        return object;
    }
    decodeSync() {
        // eslint-disable-next-line no-constant-condition
        DECODE: while (true) {
            const headByte = this.readU8();
            let object;
            if (headByte >= 0xe0) {
                // negative fixint (111x xxxx) 0xe0 - 0xff
                object = headByte - 0x100;
            }
            else if (headByte < 0xc0) {
                if (headByte < 0x80) {
                    // positive fixint (0xxx xxxx) 0x00 - 0x7f
                    object = headByte;
                }
                else if (headByte < 0x90) {
                    // fixmap (1000 xxxx) 0x80 - 0x8f
                    const size = headByte - 0x80;
                    if (size !== 0) {
                        this.pushMapState(size);
                        continue DECODE;
                    }
                    else {
                        object = {};
                    }
                }
                else {
                    // fixarray (1001 xxxx) 0x90 - 0x9f
                    const size = headByte - 0x90;
                    if (size !== 0) {
                        this.pushArrayState(size);
                        continue DECODE;
                    }
                    else {
                        object = [];
                    }
                }
            }
            else if (headByte === 0xc0) {
                // nil
                object = null;
            }
            else if (headByte === 0xc2) {
                // false
                object = false;
            }
            else if (headByte === 0xc3) {
                // true
                object = true;
            }
            else if (headByte === 0xca) {
                // float 32
                object = this.readF32();
            }
            else if (headByte === 0xcb) {
                // float 64
                object = this.readF64();
            }
            else if (headByte === 0xcc) {
                // uint 8
                object = this.readU8();
            }
            else if (headByte === 0xcd) {
                // uint 16
                object = this.readU16();
            }
            else if (headByte === 0xce) {
                // uint 32
                object = this.readU32();
            }
            else if (headByte === 0xcf) {
                // uint 64
                object = this.readU64();
            }
            else if (headByte === 0xd0) {
                // int 8
                object = this.readI8();
            }
            else if (headByte === 0xd1) {
                // int 16
                object = this.readI16();
            }
            else if (headByte === 0xd2) {
                // int 32
                object = this.readI32();
            }
            else if (headByte === 0xd3) {
                // int 64
                object = this.readI64();
            }
            else if (headByte === 0xdc) {
                // array 16
                const size = this.readU16();
                this.pushArrayState(size);
                continue DECODE;
            }
            else if (headByte === 0xdd) {
                // array 32
                const size = this.readU32();
                this.pushArrayState(size);
                continue DECODE;
            }
            else if (headByte === 0xde) {
                // map 16
                const size = this.readU16();
                this.pushMapState(size);
                continue DECODE;
            }
            else if (headByte === 0xdf) {
                // map 32
                const size = this.readU32();
                this.pushMapState(size);
                continue DECODE;
            }
            else if (headByte === 0xc4) {
                // bin 8
                const size = this.lookU8();
                object = this.decodeUtf8String(size, 1);
            }
            else if (headByte === 0xc5) {
                // bin 16
                const size = this.lookU16();
                object = this.decodeUtf8String(size, 2);
            }
            else if (headByte === 0xc6) {
                // bin 32
                const size = this.lookU32();
                object = this.decodeUtf8String(size, 4);
            }
            else if (headByte === 0xd4) {
                // fixext 1
                object = this.decodeExtension(1, 0);
            }
            else if (headByte === 0xd5) {
                // fixext 2
                object = this.decodeExtension(2, 0);
            }
            else if (headByte === 0xd6) {
                // fixext 4
                object = this.decodeExtension(4, 0);
            }
            else if (headByte === 0xd7) {
                // fixext 8
                object = this.decodeExtension(8, 0);
            }
            else if (headByte === 0xd8) {
                // fixext 16
                object = this.decodeExtension(16, 0);
            }
            else if (headByte === 0xc7) {
                // ext 8
                const size = this.lookU8();
                object = this.decodeExtension(size, 1);
            }
            else if (headByte === 0xc8) {
                // ext 16
                const size = this.lookU16();
                object = this.decodeExtension(size, 2);
            }
            else if (headByte === 0xc9) {
                // ext 32
                const size = this.lookU32();
                object = this.decodeExtension(size, 4);
            }
            else {
                throw new Error(`Unrecognized type byte: ${prettyByte(headByte)}`);
            }
            const stack = this.stack;
            while (stack.length > 0) {
                // arrays and maps
                const state = stack[stack.length - 1];
                if (state.type === 0 /* State.ARRAY */) {
                    state.array[state.position] = object;
                    state.position++;
                    if (state.position === state.size) {
                        stack.pop();
                        object = state.array;
                    }
                    else {
                        continue DECODE;
                    }
                }
                else if (state.type === 1 /* State.MAP_KEY */) {
                    // Map key
                    state.keyEnd = this.pos;
                    state.key = object;
                    state.type = 2 /* State.MAP_VALUE */;
                    continue DECODE;
                }
                else {
                    // Map value
                    if (!isValidMapKeyType(state.key)) {
                        const serializedKey = fromByteArray(this.bytes.slice(state.keyStart, state.keyEnd));
                        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                        state.map[serializedKey] = { _key: state.key, _value: object };
                    }
                    else {
                        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                        state.map[state.key] = object;
                    }
                    state.readCount++;
                    if (state.readCount === state.size) {
                        // This is the last value, end of map
                        stack.pop();
                        object = state.map;
                    }
                    else {
                        state.key = null;
                        state.type = 1 /* State.MAP_KEY */;
                        stack[stack.length - 1].keyStart = this.pos;
                        continue DECODE;
                    }
                }
            }
            return object;
        }
    }
    pushMapState(size) {
        if (size > this.maxMapLength) {
            throw new Error(`Max length exceeded: map length (${size}) > maxMapLengthLength (${this.maxMapLength})`);
        }
        this.stack.push({
            type: 1 /* State.MAP_KEY */,
            size,
            keyStart: this.pos,
            keyEnd: this.pos,
            key: null,
            serializedKey: null,
            readCount: 0,
            map: {},
        });
    }
    pushArrayState(size) {
        if (size > this.maxArrayLength) {
            throw new Error(`Max length exceeded: array length (${size}) > maxArrayLength (${this.maxArrayLength})`);
        }
        this.stack.push({
            type: 0 /* State.ARRAY */,
            size,
            keyStart: 1,
            array: new Array(size),
            position: 0,
        });
    }
    decodeUtf8String(byteLength, headerOffset) {
        var _a;
        if (byteLength > this.maxStrLength) {
            throw new Error('Max length exceeded: ' +
                `UTF-8 byte length (${byteLength}) > maxStrLength (${this.maxStrLength})`);
        }
        if (this.bytes.byteLength < this.pos + headerOffset + byteLength) {
            throw new RangeError('Insufficient data');
        }
        const offset = this.pos + headerOffset;
        let object;
        if (this.stateIsMapKey() && ((_a = this.cachedKeyDecoder) === null || _a === void 0 ? void 0 : _a.canBeCached(byteLength))) {
            object = this.cachedKeyDecoder.decode(this.bytes, offset, byteLength);
        }
        else if (TEXT_ENCODING_AVAILABLE && byteLength > TEXT_DECODER_THRESHOLD) {
            object = utf8DecodeTD(this.bytes, offset, byteLength);
        }
        else {
            object = utf8DecodeJs(this.bytes, offset, byteLength);
        }
        this.pos += headerOffset + byteLength;
        return object;
    }
    stateIsMapKey() {
        if (this.stack.length > 0) {
            const state = this.stack[this.stack.length - 1];
            return state.type === 1 /* State.MAP_KEY */;
        }
        return false;
    }
    decodeBinary(byteLength, headOffset) {
        if (byteLength > this.maxBinLength) {
            throw new Error(`Max length exceeded: bin length (${byteLength}) > maxBinLength (${this.maxBinLength})`);
        }
        if (!this.hasRemaining(byteLength + headOffset)) {
            throw new RangeError('Insufficient data');
        }
        const offset = this.pos + headOffset;
        const object = this.bytes.subarray(offset, offset + byteLength);
        this.pos += headOffset + byteLength;
        return object;
    }
    decodeExtension(size, headOffset) {
        if (size > this.maxExtLength) {
            throw new Error(`Max length exceeded: ext length (${size}) > maxExtLength (${this.maxExtLength})`);
        }
        const extType = this.view.getInt8(this.pos + headOffset);
        const data = this.decodeBinary(size, headOffset + 1 /* extType */);
        return this.extensionCodec.decode(data, extType);
    }
    lookU8() {
        return this.view.getUint8(this.pos);
    }
    lookU16() {
        return this.view.getUint16(this.pos);
    }
    lookU32() {
        return this.view.getUint32(this.pos);
    }
    readU8() {
        const value = this.view.getUint8(this.pos);
        this.pos++;
        return value;
    }
    readI8() {
        const value = this.view.getInt8(this.pos);
        this.pos++;
        return value;
    }
    readU16() {
        const value = this.view.getUint16(this.pos);
        this.pos += 2;
        return value;
    }
    readI16() {
        const value = this.view.getInt16(this.pos);
        this.pos += 2;
        return value;
    }
    readU32() {
        const value = this.view.getUint32(this.pos);
        this.pos += 4;
        return value;
    }
    readI32() {
        const value = this.view.getInt32(this.pos);
        this.pos += 4;
        return value;
    }
    readU64() {
        const value = getUint64(this.view, this.pos, this.useJSBI);
        this.pos += 8;
        return value;
    }
    readI64() {
        const value = getInt64(this.view, this.pos, this.useJSBI);
        this.pos += 8;
        return value;
    }
    readF32() {
        const value = this.view.getFloat32(this.pos);
        this.pos += 4;
        return value;
    }
    readF64() {
        const value = this.view.getFloat64(this.pos);
        this.pos += 8;
        return value;
    }
}

const defaultDecodeOptions = {};
/**
 * It decodes a MessagePack-encoded buffer.
 *
 * This is a synchronous decoding function. See other variants for asynchronous decoding:
 * `decodeAsync()`, `decodeStream()`, and `decodeArrayStream()`
 */
function decode(buffer, options = defaultDecodeOptions) {
    const decoder = new Decoder(options.extensionCodec, options.useJSBI, options.maxStrLength, options.maxBinLength, options.maxArrayLength, options.maxMapLength, options.maxExtLength, options.cachedKeyDecoder);
    decoder.setBuffer(buffer); // decodeSync() requires only one buffer
    return decoder.decodeSingleSync();
}

/* eslint-disable @typescript-eslint/prefer-for-of */
function createBaseType(value) {
    const type = typeof value;
    if (type === 'number') {
        const decimals = String(value).split('.');
        if (decimals.length > 1) {
            return new Float64(value);
        }
        return new Int(value);
    }
    if (type === 'boolean') {
        return new Bool(value);
    }
    if (type === 'string') {
        return new Str(value);
    }
    if (value !== null && type === 'object') {
        let proto = value;
        while (Object.getPrototypeOf(proto) !== null) {
            proto = Object.getPrototypeOf(proto);
        }
        if (Object.getPrototypeOf(value) === proto) {
            // eslint-disable-next-line @typescript-eslint/no-use-before-define
            return createTypedMap(value);
        }
    }
    return new Nil();
}
function createTypedMap(object) {
    const map = new Map();
    const keys = Object.keys(object);
    for (let i = 0; i < keys.length; i += 1) {
        const key = keys[i];
        const value = object[key];
        const keyType = createBaseType(key);
        const valueType = createBaseType(value);
        map.set(keyType, valueType);
    }
    return map;
}

// Copyright (c) 2018, Arista Networks, Inc.
function encodePointer(codec) {
    return (data) => {
        return encode(data.value, { extensionCodec: codec });
    };
}
function decodePointer(codec) {
    return (buffer) => {
        const pointer = decode(buffer, { extensionCodec: codec });
        if (Array.isArray(pointer)) {
            return new Pointer(pointer);
        }
        return new Pointer();
    };
}
function encodeWildcard() {
    return () => {
        return new Uint8Array();
    };
}
function decodeWildcard() {
    return () => {
        return new Wildcard();
    };
}

// Copyright (c) 2018, Arista Networks, Inc.
const POINTER_TYPE = 0;
const WILDCARD_TYPE = 1;
const NeatTypes = {
    Bool,
    Float32,
    Float64,
    Int,
    Nil,
    Pointer,
    Str,
    Wildcard,
};
const Codec = new ExtensionCodec();
Codec.register({
    type: POINTER_TYPE,
    identifier: (data) => data instanceof Pointer,
    encode: encodePointer(Codec),
    decode: decodePointer(Codec),
});
Codec.register({
    type: WILDCARD_TYPE,
    identifier: (data) => data instanceof Wildcard,
    encode: encodeWildcard(),
    decode: decodeWildcard(),
});

export { Bool, Codec, ExtData, ExtensionCodec, Float32, Float64, Int, NeatTypeSerializer, NeatTypes, Nil, Pointer, Str, createBaseType, decode, encode, isJsbi, isNeatType };
